# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Handle test context.
"""

from pytation import time, __version__
from pytation.progress import Progress
from pytation.loader import SETUP_TEARDOWN_FN, ENV_EXCLUDE
from fs.zipfs import WriteZipFS
import collections
import importlib
import zipfile
import json
import os
import logging


_FILE_FMT = "%(levelname)s:%(asctime)s:%(filename)s:%(lineno)d:%(name)s:%(message)s"
_VALID_CHARS = \
    '-_. ' \
    + ''.join([chr(ord('a') + a) for a in range(26)]) \
    + ''.join([chr(ord('a') + a) for a in range(26)]) \
    + ''.join([chr(ord('0') + a) for a in range(10)])


def sanitize_filename(s):
    s = ''.join(c for c in s if c in _VALID_CHARS)
    s = s.replace(' ', '_')  # spaces to underscore
    s = s[:64]  # truncate as needed
    return s


def _time_finalize(d):
    time_end = time.now()
    time_start = time.str_to_time(d['start'])
    duration = (time_end - time_start).total_seconds()
    d['end'] = time_end.isoformat()
    d['duration'] = str(duration)
    return d


def _json_default(obj):
    return '__pyobject__'


class DictReadOnlyWrapper(collections.Mapping):

    def __init__(self, data):
        self._data = data

    def __getitem__(self, key):
        return self._data[key]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)


class Context:
    """Context for a test station.

    :param station: The Station definition, which should already be validated
        using pytation.loader.validate.
    :ivar env: The environment, which is initialized when the station starts.
        The suite and tests may modify the environment to convey information,
        but the environment is reinitialized to the station defaults at the
        start of every suite.
    :ivar do_quit: A boolean value to indicate that the station should quit.
    """

    def __init__(self, station):
        self._log = logging.getLogger('pytation')
        self._log.setLevel(logging.DEBUG)
        self._env = {}  # cache station init to restore after each suite
        self.env: dict[str: object] = station['env']
        self._station = station

        self._progress: Progress = None
        self._devices: dict[str, object] = {}  #: string to device object
        self.devices: dict[str, object] = DictReadOnlyWrapper(self._devices)  #: dict[str, object]
        self._fs = None
        self._cbk = {'progress': [], 'state': [], 'wait_for_user': [], 'prompt': []}
        self._progress_data = []
        self._progress_file = None
        self._progress_cbk = None
        self._station_log_handler = None
        self._suite_logfile = None
        self._suite_log_file_handler = None
        self._tests = []     # The list of test outputs
        self._sections = []  # list of [name, start_time]
        self._state = None
        self.do_quit: bool = False  #: Set to True to quit, thread safe quit mechanism

    def __repr__(self):
        return 'Context(name=%s)' % self._station['name']

    @property
    def state(self):
        """The current state"""
        return self._state

    @state.setter
    def state(self, s):
        """Update the state and notify all callbacks."""
        if s not in self._station['states']:
            raise RuntimeError(f'undefined state: {s}')
        self._log.info('Enter state %s', s)
        self._state = s
        state_info = self._station['states'][s]
        for fn in self._cbk['state']:
            try:
                fn(state_info)
            except Exception:
                self._log.exception('during state callback')

    def _create_file_path_as_needed(self, path):
        dirpath = os.path.dirname(path)
        if not os.path.exists(dirpath):
            self._log.info('Creating path %s', dirpath)
            os.makedirs(dirpath, exist_ok=True)

    def expand_str(self, s):
        """Expand a string substituting environment variables.

        :param s: A string suitable for python's format with {varname}.
        :return: The string "s" with all {varname} replaced using this
            context's environment variables.
        """
        return s.format(**self.env)

    def _station_log_open(self):
        path = os.path.normpath(self.path('log'))
        self._create_file_path_as_needed(path)
        file_fmt = logging.Formatter(_FILE_FMT)
        file_hnd = logging.FileHandler(filename=path)
        file_hnd.setFormatter(file_fmt)
        file_hnd.setLevel(logging.DEBUG)
        self._station_log_handler = file_hnd
        logging.getLogger().addHandler(file_hnd)

    def _station_log_close(self):
        if self._station_log_handler is not None:
            logging.getLogger().removeHandler(self._station_log_handler)
            self._station_log_handler.close()
            self._station_log_handler = None

    def path(self, key):
        value = self._station['paths'][key]
        return value.format(**self._station['paths'], **self.env)

    def device_open(self, name):
        self._log.info('device_open(%s)', name)
        d = self._station['devices'][name]
        clz = d['clz']
        if isinstance(clz, str):
            parts = clz.split('.')
            class_name = parts[-1]
            module_name = '.'.join(parts[:-1])
            module = importlib.import_module(module_name)
            clz = getattr(module, class_name)
        if isinstance(clz, type):
            device = clz()
        elif hasattr(clz, 'setup'):
            device = clz
        else:
            raise RuntimeError(f'Invalid device clz for {name}')
        device.setup(self, d['config'])
        self._devices[name] = device
        return device

    def device_close(self, name):
        self._log.info('device_close(%s)', name)
        try:
            device = self._devices.pop(name)
        except KeyError:
            self._log.warning('device_close(%s), but not found', name)
            return
        try:
            device.teardown()
        except Exception:
            self._log.exception('device_close(%s)', name)

    def _devices_open(self, lifecycle, device_list=None):
        if device_list is None:
            return
        elif device_list is True:
            device_list = list(self._station['devices'].keys())
        for name, d in self._station['devices'].items():
            if d['lifecycle'] == lifecycle and name in device_list:
                self.device_open(name)

    def _devices_close(self, lifecycle):
        for name, d in self._station['devices'].items():
            if d['lifecycle'] == lifecycle and name in self._devices:
                self.device_close(name)

    def test_run(self, d):
        """Run a test.

        :param d: The test configuration.  The following keys are used:
            - name: The name used by the suite to save data (optional)
            - fn: The test function required.
            - config: The config dict to pass to fn (optional)
        :return: 0 on success or error code.
        """
        if d is None:
            return
        result = -1
        detail = {}
        fn = d['fn']
        if 'name' in d:
            name = d['name']
        elif hasattr(fn, 'NAME'):
            name = fn.NAME
        else:
            name = fn.__name__
        d['name'] = name
        config = d.get('config', {})
        fname = sanitize_filename(name)
        self._log.info('--- TEST START %s --- ', name)

        try:
            self._devices_open('test', d['devices'])

            if self._fs is not None and name not in SETUP_TEARDOWN_FN:
                config['fs'] = self._fs.makedir(fname)

            for d in d['devices']:
                if d not in self._devices:
                    raise RuntimeError(f'required device {d} not found')

            with self.section(name):
                result = fn(self, config)
                if result is None:
                    result = 0
                elif not isinstance(result, int):
                    result, detail = result
        except Exception:
            self._log.exception(f'While running test {name}')
        finally:
            self._devices_close('test')
            self._log.info('--- TEST DONE %s with status %s --- ', name, result)
            test = {
                'name': name,
                'result': result,
                'detail': detail
            }
            self._tests.append(test)

        for device_name, device in self._devices.items():
            try:
                device.restore()
            except Exception:
                self._log.exception('Device restore for %s', device_name)
        return result

    def _progress_exists(self):
        path = os.path.normpath(self.path('progress'))
        return os.path.isfile(path)

    def _progress_open(self):
        path = os.path.normpath(self.path('progress'))
        self._create_file_path_as_needed(path)
        if os.path.isfile(path):
            with open(path, 'r', encoding='utf-8') as f:
                self._progress = Progress(f.read())

    def _progress_save(self):
        path = os.path.normpath(self.path('progress'))
        self._create_file_path_as_needed(path)
        with open(path, 'w', encoding='utf-8') as f:
            f.write(''.join(self._progress_data))
        self._progress = None  # trigger reload

    def station_start(self):
        """Start the test station.

        :see: station_run()
        :note: Included in station_run().
        """
        self._station_log_open()
        self._log.info('pytation version = %s', __version__)
        self._devices_open('station', True)
        self.test_run(self._station.get('station_setup'))
        self._env = dict(self.env)

    def station_stop(self):
        """Stop the test station.

        :see: station_run()
        :note: Included in station_run().
        """
        self.test_run(self._station.get('station_teardown'))
        self._devices_close('station')
        self._station_log_close()

    def station_run(self, count=None):
        """Run the test suite using this station.

        :param count: The number of times to run the test suite.
            None (default) runs indefinitely.
        """
        self.station_start()
        c = 0
        try:
            while count is None or c < count:
                if self.do_quit:
                    break
                self.suite_run()
                c += 1
        except KeyboardInterrupt:
            self._log.info('KeyboardInterrupt stopped station')
        finally:
            self.station_stop()

    def _suite_time_update(self):
        t = time.now()
        self.env['suite_timestamp'] = t
        self.env['suite_timestr'] = time.time_to_filename(t)
        self.env['suite_isostr'] = time.time_to_isostr(t)

    def _suite_file_open(self):
        path = os.path.normpath(self.path('output'))
        self._log.info('suite file path = %s', path)
        self._create_file_path_as_needed(path)
        self._fs = WriteZipFS(file=path,
                              compression=zipfile.ZIP_STORED,
                              temp_fs='temp://pytation')
        self._station['env'] = dict([(key, value) for key, value in self.env.items() if key not in ENV_EXCLUDE])
        with self._fs.open('station.json', 'wt') as f:
            json.dump(self._station, f, indent=2, default=_json_default)
        self._station['env'] = {}

        # configure logging to ZIP file
        self._suite_logfile = self._fs.open('log.txt', 'wt')
        ch = logging.StreamHandler(self._suite_logfile)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s: %(message)s')
        ch.setFormatter(formatter)
        logging.getLogger().addHandler(ch)
        self._suite_log_file_handler = ch

    def _suite_start(self):
        self.env = dict(self._env)  # restore environment
        self._progress_update(0.0)
        rc = self.test_run(self._station.get('suite_setup'))  # exclude from "progress" and logging
        if rc:
            return rc
        self._suite_time_update()
        self._progress_data = []
        if self._progress is None:
            self._progress_open()
        self._suite_file_open()
        self._devices_open('suite', True)
        self._progress_file = self._fs.open('progress.csv', 'wt')
        self.section_enter('s')
        return 0

    def _progress_file_close(self):
        if self._progress_file:
            self._progress_file.close()
            self._progress_file = None
        if self._progress is None:
            self._progress_save()

    def _suite_stop(self):
        # Progress complete at this stage
        self.section_exit('s')
        self._progress_file_close()
        self._devices_close('suite')
        self._progress_update(1.0)
        self.test_run(self._station.get('suite_teardown'))
        self._log.info('*** %s ***', 'FAIL' if self.result else 'PASS')
        with self._fs.open('tests.json', 'wt') as f:
            json.dump(self._tests, f, indent=2, default=_json_default)
        if self._suite_log_file_handler:
            logging.getLogger().removeHandler(self._suite_log_file_handler)
            self._suite_log_file_handler.close()
            self._suite_log_file_handler = None
        if self._suite_logfile:
            self._suite_logfile.close()
            self._suite_logfile = None

        self._log.info('Writing zip file (may take a while)')
        self._fs.close()
        self._fs = None

    def suite_run(self):
        rc = self._suite_start()
        if rc:
            return rc
        try:
            for d in self._station['tests']:
                if self.do_quit:
                    test = {'name': 'quit', 'result': 1, 'detail': {}}
                    self._tests.append(test)
                    return 1
                self.test_run(self._station.get('test_setup'))
                result = self.test_run(d)
                self.test_run(self._station.get('test_teardown'))
                if result:
                    self.env['error_count'] += 1
                    if self.env['error_count'] >= self.env['error_count_to_halt']:
                        self._log.info('Halting due to %d errors', self.env['error_count'])
                        break
        finally:
            self._suite_stop()
        return self.result

    @property
    def result(self):
        """Get the test result.

        :return: 0 on success or error code on failure.
        """
        for test in self._tests:
            if test['result']:
                return test['result']
        return 0

    def result_str(self):
        s = []
        rv = 0
        s.append('Test results:')
        for test in self._tests:
            s.append('    %s: %s' % (test['name'], test['result']))
            if test['result'] and rv == 0:
                rv = test['result']
        s.append('*** FAIL ***' if rv else '*** PASS ***')
        return '\n'.join(s)

    def section(self, name):
        """Create a new test section as a context manager.

        :param name: The name for the subsection.
        :return: The new subsection for use in a "with" statement

        Example:

            with self.section('firmware_update'):
                do_firmware_update()
        """
        return Section(self, name)

    @property
    def section_name(self):
        return '.'.join([x[0] for x in self._sections])

    def section_enter(self, name):
        """Create a new test section.

        :param name: The name for the subsection.
        :see: section()

        Call section_exit() when done.  Alternatively, consider using
        section().
        """
        t_start = time.now()
        self._sections.append([name, t_start])
        self._log.info('%s start', self.section_name)
        self.progress('__enter__')

    def section_exit(self, name=None):
        self.progress('__exit__')
        if not len(self._sections):
            raise RuntimeError('section_stop with no section')
        section_name = self.section_name
        s_name, start_time = self._sections.pop()
        if name is not None and s_name != name:
            raise RuntimeError(f'section_stop name mismatch: {name} != {s_name}')
        stop_time = time.now()
        duration = stop_time - start_time
        self._log.info('%s: done, duration=%.3f seconds', section_name, duration)

    def _progress_update(self, progress):
        """Inform callbacks about total suite progress.

        :param progress: The total suite progress as a fract from
            0.0 (starting) and 1.0 (done).
        """
        for fn in self._cbk['progress']:
            try:
                fn(progress)
            except Exception:
                self._log.exception('during callback')

    def progress(self, progress):
        """Signal a progress step.

        :param progress: The progress which is one of:
            * A fractional floating point value between 0.0 (starting) and 1.0 (done)
            * An arbitrary string event name
        """
        if self._progress_file is None:
            return
        t = time.now() - self.env['suite_timestamp']
        section_name = self.section_name
        s = '%.3f,%s,%r\n' % (t, section_name, progress)
        self._progress_data.append(s)
        self._progress_file.write(s)
        self._log.debug('%s: %s', section_name, progress)
        if self._progress is not None:
            progress_total = self._progress.lookup(section_name, progress)
            if isinstance(progress_total, float):
                self._progress_update(progress_total)

    def wait_for_user(self):
        self.progress('__wait_enter__ wait_for_user')
        try:
            for fn in self._cbk['wait_for_user']:
                if self.do_quit:
                    raise KeyboardInterrupt('do_quit signaled')
                fn()
        finally:
            self.progress('__wait_exit__ wait_for_user')

    def prompt(self, prompt_str):
        prompt_str = str(prompt_str)
        p = f'prompt({prompt_str})'
        self.progress('__wait_enter__ ' + p)
        try:
            while True:
                for fn in self._cbk['prompt']:
                    if self.do_quit:
                        raise KeyboardInterrupt('do_quit signaled')
                    result_str = fn(prompt_str)
                    if result_str is not None:
                        self._log.info('prompt(%s) -> %s', prompt_str, result_str)
                        return result_str
        finally:
            self.progress('__wait_exit__ ' + p)

    def callback_register(self, name, cbk):
        """Register a function to call on an event.

        :param name: The callback type name, which is one of:
            [progress, state, wait_for_user]
        :param cbk: The function to call as needed.  The exact function
            prototype depends upon the name:

            - progress(progress) -> ignored:
              - progress: The progress value which is either a float
                fraction from 0.0 (start) to 1.0 (complete).
            - state(state_info) -> ignored
              - state_info: dict containing the state information from the
                station definition.  Use state_info['name'] to get the
                new state name.
            - wait_for_user() -> None
              This callback may block, but in multithreaded implementations
              it should periodically check self.do_quit.  If do_quit is
              signalled, then the callback should raise KeyboardInterrupt.
              The recommended polling interval is 10 milliseconds to
              keep the thread responsive to quit.
            - prompt(prompt_str) -> str
              - prompt_str: The string to display to the user
              - returns the string entered by the user or None on error.
        :raise KeyError: if name is not valid
        """
        self._cbk[name].append(cbk)

    def callback_unregister(self, name, cbk):
        """Remove a previously registered callback function.

         :param name: The callback type name.
         :param cbk: The function previous registered using callback_register().
         :raise KeyError: if name is not valid.
         """
        self._cbk[name] = [fn for fn in self._cbk[name] if fn != cbk]


class Section:

    def __init__(self, context, name):
        self._context = context
        self._name = name

    def __str__(self):
        return f'Section({self._name})'

    def __enter__(self):
        self._context.section_enter(self._name)
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self._context.section_exit(self._name)

    def section(self, name):
        """Create a new test subsection.

        :param name: The name for the subsection.
        :return: The new subsection for use in a "with" statement
        """
        return self._context.section(name)

    def progress(self, progress):
        """Signal a progress step.

        :param progress: The progress which is one of:
            * A fractional floating point value between 0.0 (starting) and 1.0 (done)
            * An arbitrary string event name
        """
        return self._parent.progress(progress)



# Copyright 2019-2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging


def parse_lines(txt):
    """Parse the text into the line fields

    :param txt: The text to parse.
    :return: The list with an entry for each valid line, which
        is also a list of [time, identifier, value].
    """
    offset = 0.0
    wait_start = None

    lines = []
    line = ''
    try:
        for idx, line in enumerate(txt.split('\n')):
            if not len(line) or line.startswith('#'):
                continue
            t, identifier, value = line.split(',', 2)
            t = float(t) - offset
            if value[0] == "'":
                value = value[1:-1]
                if value.startswith('__wait_enter__'):
                    wait_start = t
                elif value.startswith('__wait_exit__') and wait_start is not None:
                    t, offset = wait_start, t - wait_start
                    wait_start = None
            else:
                value = float(value)
            if wait_start is not None:
                t = wait_start
            lines.append([t, identifier, value])
    except Exception:
        logging.getLogger(__name__).error('parse failed on line %d: %s', idx + 1, line)
        raise
    return lines


def parse(txt):
    """Parse the progress log to a lookup table.

    :param txt: The text to parse.
    :return: The dict mapping 'identifier.value' to the fraction done.
    """
    lines = parse_lines(txt)
    tbl = {}
    time_end = lines[-1][0]
    for idx in range(len(lines)):
        t, identifier, value = lines[idx]
        tbl['%s.%s' % (identifier, value)] = t / time_end
    return tbl


def lookup(tbl, identifier, value):
    if isinstance(value, float):
        s_enter = tbl.get(identifier + '.__enter__')
        s_exit = tbl.get(identifier + '.__exit__')
        if s_enter is not None and s_exit is not None:
            return (s_exit - s_enter) * value + s_enter
    else:
        s = '%s.%s' % (identifier, value)
        return tbl.get(s)


class Progress:

    def __init__(self, txt):
        self.tbl = parse(txt)

    def lookup(self, identifier, value):
        return lookup(self.tbl, identifier, value)



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The API for packages using this manufacturing test framework."""


from . import Context


def declare_test(devices: list[str] = None):
    """Manufacturing test decorator to annotate required devices.

    :param devices: The list of expected devices, using the same
        device names used in the station definition.
    :return: The decorated test function(context, config).
    """
    def decorator_repeat(func):
        if devices:
            func.DEVICES = devices
        return func
    return decorator_repeat


class Device:
    """A single connected device, instrument, or sensor."""

    NAME = ''
    """The user-meaningful, descriptive test name"""

    def setup(self, context: Context, config: dict[str, object]):
        """Open and initialize the device.

        :param context: The test station context.
        :param config: Optional additional device configuration options.
        """
        raise NotImplementedError("Device.setup")

    def restore(self):
        """Restore default settings for the device.

        This function is called after each test to ensure that the
        next test starts from a known condition.
        """
        raise NotImplementedError("Device.setup")

    def teardown(self):
        """Finalize and close the device."""
        raise NotImplementedError("Device.teardown")



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pytation import Context
import os
import sys


# https://stackoverflow.com/questions/983354/how-to-make-a-script-wait-for-a-pressed-key
def wait_key():
    ''' Wait for a key press on the console and return it. '''
    result = None
    if os.name == 'nt':
        import msvcrt
        ch = msvcrt.getch()
        if ch in [b'\x03', b'\x1b']:
            raise KeyboardInterrupt('user interrupt')
        return ch
    else:
        import termios
        fd = sys.stdin.fileno()

        oldterm = termios.tcgetattr(fd)
        newattr = termios.tcgetattr(fd)
        newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
        termios.tcsetattr(fd, termios.TCSANOW, newattr)

        try:
            result = sys.stdin.read(1)
        except IOError:
            pass
        finally:
            termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)

    return result



class CliStation:

    def __init__(self, station):
        self._context = Context(station)
        self._context.callback_register('progress', self._on_progress_cbk)
        self._context.callback_register('state', self._on_state_cbk)
        self._context.callback_register('wait_for_user', self._on_wait_for_user_cbk)
        self._context.callback_register('prompt', self._on_prompt_cbk)
        self.wait_for_user = False
        self.prompt_result_str = None

    def _on_progress_cbk(self, progress):
        pass

    def _on_state_cbk(self, state):
        txt = state['html']
        print(f'State => {txt}')

    def _on_wait_for_user_cbk(self):
        print('Press a key...')
        wait_key()

    def _on_prompt_cbk(self, prompt_str):
        return input(prompt_str + '> ')

    def run(self, count=None):
        self._context.station_run(count=count)
        return 0



# Copyright 2019-2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import shutil
import subprocess


def rcc_path():
    import PySide6
    # As of PySide2 5.15.0, the pyside2-rcc executable ignores the --binary flag
    path = os.path.dirname(PySide6.__file__)
    fname = [n for n in os.listdir(path) if n.startswith('rcc')]
    if len(fname) != 1:
        raise ValueError('Could not find rcc executable')
    return os.path.join(path, fname[0])


def convert_rcc(path):
    """Convert Qt resources definitions (qrc) to binary resource files (rcc).

    :param path: The path to walk for qrc files.
    :return: The list of rcc files created, often used to construct
        a .gitignore file.
    """
    targets = []
    cmd = rcc_path()
    for root, d_names, f_names in os.walk(path):
        for source in f_names:
            _, ext = os.path.splitext(source)
            if ext == '.qrc':
                src = os.path.join(root, source)
                target = os.path.join(os.path.dirname(root), source)
                target = os.path.splitext(target)[0] + '.rcc'
                print(f'Generate {target}')
                rc = subprocess.run([cmd, src, '--binary', '--threshold', '33', '-o', target])
                if rc.returncode:
                    raise RuntimeError('failed on .qrc file')
                targets.append(target)
            else:
                continue
    return targets


def convert_ui(path, resource_import=None):
    """Convert Qt UI definitions (ui) to binary files (py).

    :param path: The path to walk for ui files.
    :param resource_import: The desired resource import python code.
    :return: The list of py files created, often used to construct
        a .gitignore file.
    """
    targets = []
    uic_path = shutil.which('pyside6-uic')
    for root, d_names, f_names in os.walk(path):
        for source in f_names:
            source = os.path.join(root, source)
            source_base, ext = os.path.splitext(source)
            if ext == '.ui':
                target = source_base + '.py'
                print(f'Generate {target}')
                rc = subprocess.run([uic_path, source], stdout=subprocess.PIPE)
                s = rc.stdout.replace(b'\r\n', b'\n').decode('utf-8')
                if resource_import is not None:
                    s = s.replace('\nimport resources_rc\n', resource_import)
                with open(target, 'w', encoding='utf-8') as ftarget:
                    ftarget.write(s)
                targets.append(target)
    return targets


def update_inno_iss(iss_path, version):
    """Update an InnoSetup file with the current version.

    :param iss_path: The path to the InnoSetup ".iss" file.
    :param version: The major.minor.patch version string.
    """
    with open(iss_path, 'r', encoding='utf-8') as fv:
        lines = fv.readlines()
    version_underscore = version.replace('.', '_')
    for idx, line in enumerate(lines):
        if line.startswith('#define MyAppVersionUnderscores'):
            lines[idx] = f'#define MyAppVersionUnderscores "{version_underscore}"\n'
        elif line.startswith('#define MyAppVersion'):
            lines[idx] = f'#define MyAppVersion "{version}"\n'
    with open(iss_path, 'w', encoding='utf-8') as fv:
        fv.write(''.join(lines))



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pytation import time
import argparse
import importlib
import os


_LOG_PATH_DEFAULT = '{base_path}/{station}/log/{station_timestr}_{process_id}.log'
_OUTPUT_PATH_DEFAULT = '{base_path}/{station}/data/{suite_timestr}.zip'
_PROGRESS_PATH_DEFAULT = '{base_path}/{station}/progress.csv'
_DEVICE_LIFECYCLE = ['station', 'suite', 'test', 'manual']  # defaults to 'station'
SETUP_TEARDOWN_FN = [
    'station_setup', 'station_teardown',
    'suite_setup', 'suite_teardown',
    'test_setup', 'test_teardown',
]
ENV_EXCLUDE = ['error_count', 'suite_timestamp', 'suite_timestr', 'suite_isostr']
ENV_DEFAULTS = {
    'error_count_to_halt': 1,
}


def parser_config(p: argparse.ArgumentParser, station=None):
    """Add station definition to an argparse.

    :param p: The argument parser instance, which will be modified
        in place.
    :param station: If provided, use the default station value with
        an optional '--station' override.  None (default) adds the station
        as a fixed position argument.
    """
    if station is not None:
        p.add_argument('--station',
                       default=station,
                       help='The fully-qualified station definition')
    else:
        p.add_argument('station',
                       help='The fully-qualified station definition')
    p.add_argument('--exclude',
                   help='The comma-separated list of tests to exlude.  Defaults to "".')
    p.add_argument('--include',
                   help='The comma-separated list of tests to include.  Defaults to all available tests.')


def _states_validate(states):
    # Shallow copy and set name field
    d = {}
    for name, s in states.items():
        s = dict(s)
        s['name'] = name
        d[name] = s
    return d


def _test_validate(test):
    if test is None:
        return None
    t = dict(test)
    fn = t['fn']
    if isinstance(fn, str):
        parts = fn.split('.')
        fn_name = parts[-1]
        module_name = '.'.join(parts[:-1])
        module = importlib.import_module(module_name)
        fn = getattr(module, fn_name)
        t['fn'] = fn

    t.setdefault('name', getattr(fn, 'NAME', getattr(fn, '__name__', '__unknown__')))
    t.setdefault('config', {})
    if 'devices' not in t:
        t['devices'] = getattr(fn, 'DEVICES', [])
    return t


def _tests_validate(test_list):
    d = []
    names = {}
    for t in test_list:
        t = _test_validate(t)
        name = t['name']
        if name in names:
            raise ValueError(f'Duplicate test name: {name}')
        d.append(t)
        names[name] = t
    return d


def _devices_validate(devices_list):
    """Convert self._station['devices'] from list of defs to dict name:def."""
    devices_map = {}
    for d in devices_list:
        d = dict(d)
        clz = d['clz']
        if 'name' in d:
            name = d['name']
        elif hasattr(clz, 'NAME'):
            name = clz.NAME
        else:
            name = clz.__name__
        d['name'] = name
        d.setdefault('lifecycle', 'station')
        d.setdefault('config', {})
        if d['lifecycle'] not in _DEVICE_LIFECYCLE:
            raise ValueError(f'invalid device lifecycle {d["lifecycle"]} for {name}')

        if name in devices_map:
            raise ValueError('Duplicate device name: %s', name)
        devices_map[name] = d
    return devices_map


def validate(station):
    """Validate the station and fully populate optional fields.

    :param station: The station data structure.
    :return: The station modified in place.
    """
    s = {}
    s['name'] = station['name']
    s['full_name'] = station.get('full_name', station['name'])

    # Construct the environment
    station_start_time = time.now()
    env = {}
    env_no_override = {
        'station': station['name'],
        'process_id': os.getpid(),
        'error_count': 0,

        'station_timestamp': station_start_time,
        'station_timestr': time.time_to_filename(station_start_time),
        'station_isostr': time.time_to_isostr(station_start_time),

        # updated at the start of each suite
        'suite_timestamp': 0,
        'suite_timestr': time.time_to_filename(0),
        'suite_isostr': time.time_to_isostr(0),
    }
    env.update(station.get('env', {}))
    env.update(env_no_override)
    for key, value in ENV_DEFAULTS.items():
        env.setdefault(key, value)
    s['env'] = env

    # Construct the station
    paths = station.get('paths', {})
    paths.setdefault('base_path', os.path.join(os.path.expanduser('~'), 'pytation'))
    paths.setdefault('log', _LOG_PATH_DEFAULT)
    paths.setdefault('output', _OUTPUT_PATH_DEFAULT)
    paths.setdefault('progress', _PROGRESS_PATH_DEFAULT)
    s['paths'] = paths
    s['states'] = _states_validate(station.get('states', {}))
    s['tests'] = _tests_validate(station['tests'])
    s['devices'] = _devices_validate(station['devices'])
    for k in SETUP_TEARDOWN_FN:
        s[k] = _test_validate(station.get(k, None))
    s['gui_resources'] = station.get('gui_resources', [])
    
    return s


def load(args):
    """Load a station from the command-line arguments.

    :param args: The command-line arguments.
    :return: The station, which is also fully validated.
    :see: parser_config()
    :see: validate()
    """
    parts = args.station.split('.')
    def_name = parts[-1]
    module_name = '.'.join(parts[:-1])
    module = importlib.import_module(module_name)
    station = getattr(module, def_name)
    station = validate(station)

    if args.exclude is not None:
        exclude = args.exclude.split(',')
        tests = []
        for test in station['tests']:
            if test['name'] in exclude:
                exclude.remove(test['name'])
            else:
                tests.append(test)
        if len(exclude):
            raise ValueError(f'Excluded tests not found: {exclude}')
        station['tests'] = tests

    if args.include is not None:
        include = args.include.split(',')
        tests = []
        for test in station['tests']:
            if test['name'] in include:
                include.remove(test['name'])
                tests.append(test)
        if len(include):
            raise ValueError(f'Include tests not found: {include}')
        station['tests'] = tests

    return station



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import time
from pytation import Context

MYPATH = os.path.dirname(os.path.abspath(__file__))


class Eq1:
    NAME = 'Equipment Example 1'

    def __init__(self):
        self._log = logging.getLogger(__name__ + '.Eq1')

    def setup(self, context: Context, config=None):
        self._log.info('setup')

    def restore(self):
        self._log.info('restore')

    def teardown(self):
        self._log.info('teardown')


class Dut1:
    NAME = 'Device under Test Example 1'

    def __init__(self):
        self._log = logging.getLogger(__name__ + '.Dut1')

    def setup(self, context: Context, config=None):
        self._log.info('setup')

    def restore(self):
        self._log.info('restore')

    def teardown(self):
        self._log.info('teardown')


def suite_setup(context, config=None):
    context.state = 'wait_for_dut'
    context.wait_for_user()
    context.state = 'in_progress'


def suite_teardown(context, config=None):
    if context.result == 0:
        context.state = 'pass'
    else:
        context.state = 'fail'
    context.wait_for_user()


def enter_serial_number(context, config=None):
    serial_number = context.prompt('Enter serial number')
    context.env['name'] = serial_number
    context.env['serial_number'] = serial_number
    return 0, {'serial_number': serial_number}


def test1(context, config=None):
    config = {} if config is None else config
    _log = logging.getLogger(__name__ + '.test1')
    mode = config.get('mode', 'normal')
    delay = float(config.get('delay', 0.01))
    count = int(config.get('count', 50))
    _log.info('start: mode=%s', mode)
    for i in range(count):
        time.sleep(delay)
        context.progress((i + 1) / count)
    _log.info('stop')
    return 0, {'hello': 'world'}


STATION = {
    'name': 'simple',
    'full_name': 'Simple test station example',
    'env': {},
    'suite_setup': {'fn': suite_setup, 'config': {}},
    'suite_teardown': {'fn': suite_teardown, 'config': {}},
    'states': {
        'initialize': {
            'pixmap': ':/station/initialize.jpg',
            'style': 'QLabel { background-color : yellow; color : black; font-size : 12pt; }',
            'html': '<html><body><h1>Connect Equipment</h1></body></html>',
        },
        'wait_for_dut': {
            'pixmap': ':/station/wait_for_dut.jpg',
            'style': 'QLabel { background-color : #8080ff; color : black; font-size : 12pt; }',
            'html': '<html><body><h1>Connect Device Under Test and press any key</h1></body></html>',
        },
        'in_progress': {
            'pixmap': ':/station/in_progress.jpg',
            'style': 'QLabel { background-color : white; color : black; font-size : 12pt; }',
            'html': '<html><body><h1>Test in Progress</h1></body></html>',
        },
        'pass': {
            'pixmap': ':/station/pass.jpg',
            'style': 'QLabel { background-color : green; color : black; font-size : 12pt; }',
            'html': '<html><body><h1>PASS</h1><p>Disconnect the device and press any key.</p></body></html>',
        },
        'fail': {
            'pixmap': ':/station/fail.jpg',
            'style': 'QLabel { background-color : red; color : black; font-size : 12pt; }',
            'html': '<html><body><h1>FAILED</h1><p>Disconnect the device and press any key.</p></body></html>',
        },
        'abort': {
            'pixmap': ':/station/abort.jpg',
            'style': 'QLabel { background-color : red; color : black; font-size : 12pt; }',
            'html': '<html><body><h1>ABORT</h1><p>Internal error - close & restart</p></body></html>',
        },
    },
    'tests': [
        {'fn': enter_serial_number},
        {'fn': test1},
        {'name': 'long_iter', 'fn': test1, 'config': {'delay': 0.5, 'count': 4}},
    ],
    'devices': [
        {'name': 'eq1', 'clz': Eq1},
        {'name': 'dut', 'clz': Dut1, 'lifecycle': 'suite', 'config': {'mode': 'test'}},
    ],
    'gui_resources': [['pytation_examples', 'pytation_examples.rcc']]  # list of [package, resource]
}



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the Context class.
"""

import unittest
from unittest.mock import Mock
from pytation import Context, declare_test
from pytation.loader import validate


@declare_test(['eq2'])
def needs_eq2(context, config):
    return 0


def call_wait_for_user(context, config):
    context.wait_for_user()


def call_prompt(context, config):
    s = context.prompt('Hello?')
    return 0, {'user_data': s}


def state_in_progress(context, config):
    context.state = 'in_progress'


class TestContext(unittest.TestCase):

    def _station1(self, name, skip_validate=False):
        self.test1 = Mock()
        self.test1.return_value = 0
        self.test1.DEVICES = ['dut', 'eq1']
        self.test2 = Mock()
        self.test2.return_value = 0
        self.test2.DEVICES = ['dut', 'eq1']
        self.eq1 = Mock(['setup', 'restore', 'teardown'])
        self.dut = Mock(['setup', 'restore', 'teardown'])
        station = {
            'name': name,
            'full_name': 'Station 1 for unit test',
            'states': {
                'initialize': {},
                'wait_for_dut': {},
                'in_progress': {},
                'pass': {},
                'fail': {},
                'abort': {},
            },
            'tests': [
                {'name': 'test1', 'fn': self.test1, 'config': {}},
                {'name': 'test2', 'fn': self.test2, 'config': {}},
            ],
            'devices': [
                {'name': 'eq1', 'clz': self.eq1, 'config': {}},
                {'name': 'dut', 'clz': self.dut, 'lifecycle': 'suite', 'config': {'mode': 'test'}},
            ],
        }
        if skip_validate:
            return station
        return validate(station)

    def test_station_start_and_stop(self):
        context = Context(self._station1('test_station_start_and_stop'))
        context.station_start()
        self.eq1.setup.assert_called_once()
        context.station_stop()
        self.eq1.teardown.assert_called_once()

    def test_station_once(self):
        context = Context(self._station1('test_station_once'))
        context.station_run(count=1)
        self.eq1.setup.assert_called_once()
        self.dut.setup.assert_called_once()
        self.test1.assert_called_once()
        self.test2.assert_called_once()
        self.dut.teardown.assert_called_once()
        self.eq1.teardown.assert_called_once()

    def test_station_manual(self):
        context = Context(self._station1('test_station_manual'))
        context.station_start()
        self.assertEqual(0, context.suite_run())
        context.station_stop()

    def test_station_failure(self):
        context = Context(self._station1('test_station_failure'))
        self.test1.return_value = 1
        context.station_start()
        self.assertEqual(1, context.suite_run())
        context.station_stop()

    def test_return_values(self):
        context = Context(self._station1('test_return_values'))
        self.test1.return_value = None
        self.test2.return_value = 0, {'hello': 'world'}
        context.station_start()
        self.assertEqual(0, context.suite_run())
        context.station_stop()

    def test_with_required_device_not_present(self):
        station = self._station1('test_with_required_device_not_present', skip_validate=True)
        station['tests'][0]['fn'] = needs_eq2
        station = validate(station)
        context = Context(station)
        context.station_start()
        self.assertEqual(-1, context.suite_run())
        context.station_stop()

    def test_wait_for_user_callback(self):
        cbk = Mock()
        station = self._station1('test_wait_for_user_callback')
        station['tests'][0]['fn'] = call_wait_for_user
        context = Context(station)
        context.callback_register('wait_for_user', cbk)
        context.station_run(count=1)
        cbk.assert_called_once()

    def test_prompt(self):
        cbk = Mock()
        cbk.return_value = 'yes!'
        station = self._station1('test_wait_for_user_callback')
        station['tests'][0]['fn'] = call_prompt
        context = Context(station)
        context.callback_register('prompt', cbk)
        context.station_run(count=1)
        cbk.assert_called_once()

    def test_state(self):
        cbk = Mock()
        station = self._station1('test_state')
        station['tests'][0]['fn'] = state_in_progress
        context = Context(station)
        context.callback_register('state', cbk)
        context.station_run(count=1)
        cbk.assert_called_once()



# Copyright 2019-2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from pytation import __version__, Context
from PySide6 import QtCore, QtWidgets, QtGui
from time import sleep
import pkgutil
import queue
import threading
import logging


LOG_COLOR = {
    logging.CRITICAL: QtGui.QColor(255, 0, 0),
    logging.ERROR: QtGui.QColor(128, 0, 0),
    logging.WARNING: QtGui.QColor(255, 165, 0),
    logging.INFO: QtGui.QColor(0, 0, 0),
    logging.DEBUG: QtGui.QColor(128, 128, 128),
}


class Handler(logging.Handler):

    def __init__(self, cbk):
        logging.Handler.__init__(self)
        self._cbk = cbk

    def emit(self, record):
        self._cbk(record)


class Formatter(logging.Formatter):
    def formatException(self, ei):
        result = super(Formatter, self).formatException(ei)
        return result

    def format(self, record):
        s = super(Formatter, self).format(record)
        if record.exc_text:
            s = s.replace('\n', '')
        return s


class QResyncEvent(QtCore.QEvent):
    """An event containing a request for python message processing."""
    EVENT_TYPE = QtCore.QEvent.Type(QtCore.QEvent.registerEventType())

    def __init__(self):
        QtCore.QEvent.__init__(self, self.EVENT_TYPE)

    def __str__(self):
        return 'QResyncEvent()'

    def __len__(self):
        return 0


class StationObject:

    def __init__(self, parent, station):
        self._context = Context(station)
        self._parent = parent
        self._log = logging.getLogger('station')
        self._context.callback_register('progress', self._on_progress_cbk)
        self._context.callback_register('state', self._on_state_cbk)
        self._context.callback_register('wait_for_user', self._on_wait_for_user_cbk)
        self._context.callback_register('prompt', self._on_prompt_cbk)
        self.wait_for_user = False
        self.prompt_result_str = None

    def _on_progress_cbk(self, progress):
        self._parent.on_message({'type': 'progress', 'data': progress})

    def _on_state_cbk(self, state):
        self._parent.on_message({'type': 'state', 'data': state})

    def _on_wait_for_user_cbk(self):
        self.wait_for_user = False
        while True:
            if self._context.do_quit:
                raise KeyboardInterrupt('do_quit while wait_for_user')
            elif self.wait_for_user:
                return
            sleep(0.05)

    def _on_prompt_cbk(self, prompt_str):
        self.prompt_result_str = None
        self._parent.on_message({'type': 'prompt', 'data': prompt_str})
        while True:
            if self._context.do_quit:
                raise KeyboardInterrupt('do_quit while prompt')
            elif self.prompt_result_str is False:
                return None
            elif self.prompt_result_str is not None:
                return self.prompt_result_str
            sleep(0.05)

    def run(self):
        self._log.info('Station thread starting')
        self._context.station_run()
        self._log.info('Station thread stopping')

    @QtCore.Slot()
    def quit_request(self):
        self._context.do_quit = True


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self, app, station):
        super(MainWindow, self).__init__()
        self.setObjectName('MainWindow')
        self._app = app
        self._handler = Handler(self._on_log_cbk)
        self._formatter = Formatter('%(asctime)s %(levelname)s %(message)s', '%Y%m%d %H:%M:%S')
        self._log = logging.getLogger()
        self._station_thread = None
        self._station = None
        self._queue = queue.Queue()
        self._pixmap_bin = None

        self.resize(800, 600)

        #window_icon = station.get('window_icon')
        #if window_icon is not None:
        #    self._icon = QIcon()
        #    self._icon.addFile(window_icon, QSize(), QIcon.Normal, QIcon.Off)
        #    self.setWindowIcon(self._icon)

        self._central_widget = QtWidgets.QWidget(self)
        self._central_widget.setObjectName('_central_widget')
        self.setCentralWidget(self._central_widget)
        self._vertical_layout = QtWidgets.QVBoxLayout(self._central_widget)
        self._vertical_layout.setObjectName('_vertical_layout')

        self._frame = QtWidgets.QFrame(self._central_widget)
        self._frame.setObjectName('_frame')
        self._frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self._frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self._vertical_layout.addWidget(self._frame)
        self._horizontal_layout = QtWidgets.QHBoxLayout(self._frame)
        self._horizontal_layout.setObjectName('_horizontal_layout')

        self._image_label = QtWidgets.QLabel(self._frame)
        self._image_label.setObjectName('_image_label')
        size_policy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        size_policy.setHorizontalStretch(1)
        size_policy.setVerticalStretch(0)
        size_policy.setHeightForWidth(self._image_label.sizePolicy().hasHeightForWidth())
        self._image_label.setSizePolicy(size_policy)
        self._image_label.setAlignment(QtGui.Qt.AlignCenter)
        self._horizontal_layout.addWidget(self._image_label)

        self._stage_text = QtWidgets.QLabel(self._frame)
        self._stage_text.setObjectName('_stage_text')
        size_policy.setHeightForWidth(self._stage_text.sizePolicy().hasHeightForWidth())
        self._stage_text.setSizePolicy(size_policy)
        self._stage_text.setAlignment(QtGui.Qt.AlignCenter)
        self._horizontal_layout.addWidget(self._stage_text)

        self._logging_textEdit = QtWidgets.QTextEdit(self._central_widget)
        self._logging_textEdit.setObjectName('_logging_textEdit')
        self._logging_textEdit.setUndoRedoEnabled(False)
        self._logging_textEdit.setReadOnly(True)
        self._logging_textEdit.setTextInteractionFlags(QtGui.Qt.NoTextInteraction)
        self._vertical_layout.addWidget(self._logging_textEdit)

        self._progress_bar = QtWidgets.QProgressBar(self._central_widget)
        self._progress_bar.setObjectName('_progress_bar')
        self._progress_bar.setRange(0, 1000)
        self._vertical_layout.addWidget(self._progress_bar)

        self._logging_configure()
        self._log.info('\n**************************************' +
                       '\n* STATION=%s' +
                       '\n* VERSION=%s' +
                       '\n**************************************',
                       station['name'], __version__)
        self._log_demo()

        self._shortcut_spacebar = QtGui.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Space), self)
        self._shortcut_spacebar.activated.connect(self._on_spacebar)

        # self.showFullScreen()
        self.showMaximized()
        self._station_start(station)

    def _station_start(self, station):
        self._station = StationObject(self, station)

        self._station_thread = threading.Thread(target=self._station.run)
        self._station_thread.start()

    def _station_stop(self):
        if self._station_thread is not None:
            self._station.quit_request()
            self._station_thread.join()
            self._station = None
            self._station_thread = None

    @QtCore.Slot(float)
    def _on_progress(self, progress):
        if progress <= 0:
            self._logging_textEdit.clear()
        self._progress_bar.setValue(round(progress * 1000))

    @QtCore.Slot(float)
    def _on_state(self, state):
        pixmap = state.get('pixmap')
        style = state.get('style', 'QLabel { background-color : white; color : black; font-size : 12pt; }')
        html = state.get('html', '')
        if pixmap is None:
            self._image_label.clear()
            self._pixmap_bin = None
        else:
            try:
                pixmap_bin = QtGui.QPixmap(pixmap)
                self._log.info('image_label pixmap: %s => %s', pixmap, pixmap_bin.size())
                if pixmap_bin is None:
                    self._log.warning('QPixmap failed for resource %s', state.get('pixmap'))
                self._image_label.setPixmap(pixmap_bin)
                self._pixmap_bin = pixmap_bin
            except Exception:
                self._log.warning('Could not load image')
        if style is not None:
            self._stage_text.setStyleSheet(style)
        if html is not None:
            self._stage_text.setText(html)

    @QtCore.Slot(str)
    def _on_prompt(self, prompt_str):
        s, status = QtWidgets.QInputDialog.getText(self, "Input data", prompt_str)
        if s is None or not status:
            self._station.prompt_result_str = False
        else:
            self._station.prompt_result_str = s

    def _logging_configure(self):
        self._handler.setLevel(logging.INFO)
        self._log.addHandler(self._handler)
        self._log.setLevel(logging.DEBUG)

    def _log_demo(self):
        self._log.critical('critical')
        self._log.error('error')
        self._log.warning('warning')
        self._log.info('info')
        self._log.debug('debug')

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        key = event.key()
        if key in [QtCore.Qt.Key_Escape]:
            self.close()
        else:
            self._station.wait_for_user = True
        event.accept()

    def _on_spacebar(self):
        self._station.wait_for_user = True

    def event(self, event: QtCore.QEvent):
        if event.type() == QResyncEvent.EVENT_TYPE:
            # process resync_handler resync calls.
            event.accept()
            self._message_process()
            return True
        else:
            return super(MainWindow, self).event(event)

    def _message_process(self):
        while True:
            try:
                msg = self._queue.get(block=False)
                msg_type = msg['type']
                if msg_type == 'log':
                    self.on_logRecord(msg['data'])
                elif msg_type == 'progress':
                    self._on_progress(msg['data'])
                elif msg_type == 'state':
                    self._on_state(msg['data'])
                elif msg_type == 'prompt':
                    self._on_prompt(msg['data'])
            except queue.Empty:
                break

    def on_message(self, msg):
        self._queue.put(msg)
        event = QResyncEvent()
        QtCore.QCoreApplication.postEvent(self, event)

    def _on_log_cbk(self, record):
        msg = {
            'type': 'log',
            'data': record
        }
        self.on_message(msg)

    @QtCore.Slot(object)
    def on_logRecord(self, record: logging.LogRecord):
        txt = self._formatter.format(record)
        color = LOG_COLOR[record.levelno]
        self._logging_textEdit.setTextColor(color)
        self._logging_textEdit.append(txt)

    def closeEvent(self, event: QtGui.QCloseEvent):
        self._log.info('MainWindow.closeEvent')
        self._station_stop()
        event.accept()


def run(station):
    #if sys.platform.startswith('win'):
    #    ctypes.windll.user32.SetProcessDPIAware()
    #QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    app = QtWidgets.QApplication(sys.argv)

    resources = []
    log = logging.getLogger()
    for package, data in station['gui_resources']:
        b = pkgutil.get_data(package, data)
        log.info(f'resource {package} {data} => {len(b)} bytes')
        if b is None or not len(b):
            log.warning('Could not find resource: %s %s', package, data)
        if not QtCore.QResource.registerResourceData(b):
            log.warning('Could not register resource: %s %s', package, data)
        resources.append(b)

    try:
        ui = MainWindow(app, station)
        rc = app.exec_()
    except Exception:
        log.exception('while running MainWindow')
        raise
    return rc



__version__ = "0.1.1"

__title__ = "pytation"
__description__ = 'Create manufacturing and validation tests for hardware development.'
__url__ = 'https://github.com/jetperch/pytation'
__author__ = 'Jetperch LLC'
__author_email__ = 'joulescope-dev@jetperch.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2018-2021 Jetperch LLC'

__all__ = ['__version__', '__title__', '__description__', '__url__',
           '__author__', '__author_email__', '__license__', '__copyright__']



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Define the time operations for this package."""


import datetime
import time


def now():
    """Get the current time as seconds since the POSIX epoch (UTC)."""
    return time.time()


def time_to_filename(t=None):
    if t is None:
        t = now()
    dt = datetime.datetime.utcfromtimestamp(t)
    return dt.strftime('%Y%m%d_%H%M%S')


def time_to_isostr(t=None):
    if t is None:
        t = now()
    dt = datetime.datetime.utcfromtimestamp(t)
    return dt.isoformat()


def isostr_to_time(s):
    dt = datetime.datetime.fromisoformat(s)
    return dt.timestamp()



# Copyright 2019-2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest.mock import Mock
from pytation import loader
from pytation import declare_test, Context
import argparse


@declare_test(['dut'])
def test2(context: Context, config: dict[str, object]):
    return 0


class DutEmpty:
    def setup(self, context: Context, config=None):
        pass

    def restore(self):
        pass

    def teardown(self):
        pass


class TestLoader(unittest.TestCase):

    def _parser(self):
        p = argparse.ArgumentParser()
        loader.parser_config(p)
        return p

    def test_simple_normal(self):
        args = self._parser().parse_args(['pytation_examples.simple.STATION'])
        s = loader.load(args)
        tests = s['tests']
        self.assertEqual(3, len(tests))
        self.assertEqual(['enter_serial_number', 'test1', 'long_iter'], [t['name'] for t in tests])

    def test_exclude_one(self):
        args = self._parser().parse_args(['pytation_examples.simple.STATION', '--exclude', 'test1'])
        s = loader.load(args)
        tests = s['tests']
        self.assertEqual(['enter_serial_number', 'long_iter'], [t['name'] for t in tests])

    def test_exclude_multiple(self):
        args = self._parser().parse_args(['pytation_examples.simple.STATION', '--exclude', 'test1,enter_serial_number'])
        s = loader.load(args)
        tests = s['tests']
        self.assertEqual(['long_iter'], [t['name'] for t in tests])

    def test_exclude_invalid(self):
        args = self._parser().parse_args(['pytation_examples.simple.STATION', '--exclude', '__invalid__'])
        with self.assertRaises(ValueError):
            loader.load(args)

    def test_include_one(self):
        args = self._parser().parse_args(['pytation_examples.simple.STATION', '--include', 'test1'])
        s = loader.load(args)
        tests = s['tests']
        self.assertEqual(['test1'], [t['name'] for t in tests])

    def test_include_multiple(self):
        args = self._parser().parse_args(['pytation_examples.simple.STATION', '--include', 'test1,enter_serial_number'])
        s = loader.load(args)
        tests = s['tests']
        self.assertEqual(['enter_serial_number', 'test1'], [t['name'] for t in tests])


class TestValidator(unittest.TestCase):

    def setUp(self):
        self.test1 = Mock()
        self.test1.NAME = 'test1'
        self.test1.return_value = 0
        self.test1.DEVICES = ['dut', 'eq1']
        self.eq1 = Mock(['setup', 'restore', 'teardown'])
        self.dut = Mock(['setup', 'restore', 'teardown'])
        self.dut.NAME = 'dut'

    def test_validate_basics(self):
        station = {
            'name': 'station1',
            'tests': [
                {'fn': self.test1},
                {'fn': test2},
                {'name': 'test3', 'fn': test2},
            ],
            'devices': [
                {'name': 'eq1', 'clz': self.eq1, 'config': {}},
                {'clz': self.dut, 'lifecycle': 'suite', 'config': {'mode': 'test'}},
            ],
        }
        station = loader.validate(station)
        tests = station['tests']
        self.assertEqual('test1', tests[0]['name'])
        self.assertEqual(['dut', 'eq1'], tests[0]['devices'])
        self.assertEqual({}, tests[0]['config'])
        self.assertEqual('test2', tests[1]['name'])
        self.assertEqual('test3', tests[2]['name'])

    def test_duplicate_test_name(self):
        station = {
            'name': 'station1',
            'tests': [{'fn': test2}, {'fn': test2}],
            'devices': [{'clz': self.dut, 'lifecycle': 'suite', 'config': {'mode': 'test'}}],
        }
        with self.assertRaises(ValueError):
            loader.validate(station)

    def test_duplicate_device_name(self):
        station = {
            'name': 'station1',
            'tests': [{'fn': test2}],
            'devices': [{'clz': self.dut}, {'clz': self.dut}],
        }
        with self.assertRaises(ValueError):
            loader.validate(station)

    def test_invalid_device_lifecycle(self):
        station = {
            'name': 'station1',
            'tests': [{'fn': test2}],
            'devices': [{'clz': self.dut, 'lifecycle': 'invalid'}],
        }
        with self.assertRaises(ValueError):
            loader.validate(station)

    def test_device_name(self):
        dut2 = Mock(['setup', 'restore', 'teardown'])
        station = {
            'name': 'station1',
            'tests': [{'fn': test2}],
            'devices': [{'clz': self.dut}, {'name': 'dut2', 'clz': dut2}, {'clz': DutEmpty}],
        }
        station = loader.validate(station)
        names = [d['name'] for d in station['devices'].values()]
        self.assertEqual(['dut', 'dut2', 'DutEmpty'], names)



# Copyright 2019-2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from pytation.progress import parse, lookup, Progress


TXT1 = """\
0.033,,'__enter__'
0.040,program,'__enter__'
0.043,program,'find_lpc_bootloader'
0.046,program,'program_bootloader'
2.622,program,'wait_for_bootloader'
2.623,program,'program_personality'
2.754,program,'controller_program'
2.757,program.controller,'__enter__'
2.757,program.controller,'load firmware'
2.758,program.controller,'find bootloader'
2.761,program.controller,'open bootloader'
2.765,program.controller.firmware,'__enter__'
4.118,program.controller.firmware,'__exit__'
4.119,program.controller,'launch application'
4.121,program.controller,'__exit__'
4.122,program,'sensor_program'
4.123,program.sensor,'__enter__'
4.124,program.sensor,'load firmware'
4.125,program.sensor,'find Joulescope'
5.641,program.sensor,'open Joulescope'
6.365,program.sensor.firmware,'__enter__'
8.0,program.sensor.firmware,0.1
10.0,program.sensor.firmware,0.2
12.0,program.sensor.firmware,0.4
14.0,program.sensor.firmware,0.6
16.0,program.sensor.firmware,0.8
20.0,program.sensor.firmware,0.9
21.000,program.sensor.firmware,'__exit__'
21.030,program.sensor,'__exit__'
21.033,program,'__exit__'
21.066,,'__exit__'
"""

tbl = parse(TXT1)

TXT2 = """\
0.0,s,'__enter__'
0.1,s.hi,'__enter__'
0.5,s.hi,'hello'
1.0,s.hi,'__wait_enter__ user_input'
5.0,s.hi,'__wait_exit__ user_input'
6.0,s.hi,'world'
9.0,s.hi,'__exit__'
10.0,s,'__exit__'
"""

class TestLookup(unittest.TestCase):

    def test_root(self):
        self.assertLess(lookup(tbl, '', '__enter__'), 0.01)
        self.assertGreater(lookup(tbl, '', '__exit__'), 0.99)

    def test_exact(self):
        self.assertAlmostEqual(lookup(tbl, 'program.controller', 'find bootloader'), 0.13092186461596886)

    def test_float(self):
        s_enter = lookup(tbl, 'program.sensor.firmware', '__enter__')
        s_exit = lookup(tbl, 'program.sensor.firmware', '__exit__')
        s_mid = (s_exit + s_enter) / 2
        self.assertAlmostEqual(lookup(tbl, 'program.sensor.firmware', 0.0), s_enter)
        self.assertAlmostEqual(lookup(tbl, 'program.sensor.firmware', 1.0), s_exit)
        self.assertAlmostEqual(lookup(tbl, 'program.sensor.firmware', 0.5), s_mid)

    def test_class(self):
        c = Progress(TXT1)
        self.assertLess(c.lookup('', '__enter__'), 0.01)
        self.assertGreater(c.lookup('', '__exit__'), 0.99)
        
    def test_wait(self):
        tbl = parse(TXT2)
        self.assertAlmostEqual(lookup(tbl, 's.hi', 'world'), 1/3)
